#ifndef _LOGINT_HH
#define _LOGINT_HH


#include <string.h>
#include <iostream.h>
#include <memory.h>

class longInt
{
public:

  enum   sign
  {
    pos = 1,    // positive
    neg  // negtive
  } ;
  

  //constructors and destructor
  longInt();
  longInt(int,sign,char);
  longInt(char*);
  longInt(const longInt&);
  ~longInt();
  
  //assignment operators
  longInt& operator= (const longInt&);
  longInt& operator= (const char*);

  longInt abs() const; //absolute value
  longInt half() const; //return half of the long int
  longInt doub() const; //return bouble of the long int
  

  //Arithmetic operators
  longInt operator-() const; //unary
  friend longInt operator+ (const longInt&, const longInt&);
  friend longInt operator- (const longInt&, const longInt&);
  friend longInt operator* (const longInt&, const longInt&);
  friend longInt operator* (const longInt&, const int&);
  friend longInt operator/ (const longInt&, const longInt&);
  friend longInt operator% (const longInt&, const longInt&);
  



  //comparison operators
  friend bool operator< (const longInt&, const longInt&);
  friend bool operator<= (const longInt&, const longInt&);
  friend bool operator> (const longInt&, const longInt&);
  friend bool operator>= (const longInt&, const longInt&);
  friend bool operator== (const longInt&, const longInt&);
  

  friend ostream& operator<< (ostream&, const longInt&);
  friend istream& operator>> (istream&, longInt&);
  

  
private:
  
  char* val_; //hold the absolute value
  int size_;  //length of the absolute value
  sign sign_; //positive or negtive

  void reduce(); //move leading zeros 
  bool isDivision(const longInt&, const longInt&); //for operator /
  
};


longInt::longInt()
{
  val_ = NULL;
  size_ = 0;
  sign_ = pos;
};



longInt::longInt(int sz, sign sg, char c)
{
  size_ = sz;
  sign_ = sg;
  

  val_ = new char[sz+1];
  

  for(int i=0; i<sz; i++){
    val_[i] = c;
  }
  

};


  
//precondition: s must be a valid int (e.g. -21 or 0)
//no error checking in the code
longInt::longInt(char* s)
{

  if(NULL == s){
    val_ = NULL;
    size_ = 0;
    sign_ = pos;
  }
  else{
    if(s[0]=='-'){ //negative int
      size_ = strlen(s)-1;
      val_ = new char[size_+1];
      strcpy(val_,s+sizeof(char));
      sign_ = neg;
    }
    else{
      size_ = strlen(s);
      val_ = new char[size_+1];
      strcpy(val_,s);
      sign_ = pos;
    }
  }

  reduce();
 
}
  

longInt::longInt(const longInt& li)
{

  val_ = new char[li.size_+1];
  for(int i=0; i<li.size_; i++){
      val_[i] = li.val_[i];
  }
  
  size_ = li.size_;
  sign_ = li.sign_;

}
    

longInt::~longInt()
{
  if(val_) delete[] val_;
}



//assignment operators
longInt& longInt::operator= (const longInt& rhs)
{
  

  if(this!=&rhs){
    if(val_) delete[] val_;
    val_ = new char[rhs.size_+1];
    for(int i=0; i<rhs.size_; i++){
      val_[i] = rhs.val_[i];
    }
    
    size_ = rhs.size_;
    sign_ = rhs.sign_;
  }
  

  return *this;
}


longInt& longInt::operator= (const char* s)
{
  if(NULL == s){
    val_ = NULL;
    size_ = 0;
    sign_ = pos;
  }
  else{
    if(s[0]=='-'){ //negative int
      size_ = strlen(s)-1;
      val_ = new char[size_+1];
      strcpy(val_,s+sizeof(char));
      sign_ = neg;
    }
    else{
      size_ = strlen(s);
      val_ = new char[size_+1];
      strcpy(val_,s);
      sign_ = pos;
    }
  }

  reduce();
  
  return *this;
}




longInt longInt::abs() const 
{
  longInt result(*this);
  if (sign_ == neg) result.sign_ = pos;
  
  return result;

}


longInt longInt::half() const 
{
  longInt zero("0");
  if(*this==zero){
    return zero;
  }

  longInt result(*this);
  
  int i, v,s,rem=0;
  
  for(i=0; i<size_; i++){
    v = (int)val_[i] - (int)'0' + rem*10;
    s = v/2;
    rem = v%2;
    result.val_[i]=(char) (s+(int)'0');
  }
  
  result.reduce();
  return result;
  

}

longInt longInt::doub() const
{
  longInt two("2");
  return (*this)*two;
}




//unary
longInt longInt::operator-() const 
{ 
  longInt result(*this);
  
  if(result.sign_ == pos) result.sign_ = neg;
  else result.sign_ = pos;
  
  return result;
 
}



longInt operator+ (const longInt& left, const longInt& right)
{
  longInt result, shortInt;
  int i,s; //sum in a single digit
  int extra = 0, rem; 
  
  //cursor of result and shortInt
  int hand_short, hand_result; 
 
  if(left.sign_==right.sign_){ //same sign
    
    if(left.size_>=right.size_){
      result = left;
      shortInt = right;
    }
    else{
      result = right;
      shortInt = left;
    }
    
    
    for(i = 0; i<shortInt.size_; i++){
      hand_short = shortInt.size_-1-i;
      hand_result = result.size_-1-i;
      
      s = (int)shortInt.val_[hand_short]-(int)'0'
	+(int)result.val_[hand_result]-(int)'0' + extra;
      
      if(s>=10){
	extra = s/10;
	rem = s%10;
      }
      else{
	extra = 0;
	rem = s;
      }
      result.val_[hand_result] = (char) (rem+(int)'0');
    };
      
    if(extra>0){
      
      while(hand_result>0){
	hand_result--;
	s = (int)result.val_[hand_result]-(int)'0' + extra;
	if(s>=10){
	  extra = s/10;
	  rem = s%10;
	  result.val_[hand_result] = (char) (rem+(int)'0');
	}
	else{
	  extra = 0;
	  result.val_[hand_result] = (char) (s+(int)'0');
	  break;
	}
      }//end while
      
    } //end if (extra>0){
    
    if(extra>0){//need one more space to hold result
      result.size_++;
      char* temp = result.val_;
      result.val_ = new char[result.size_+1];
      strcpy(result.val_+sizeof(char),temp);
      result.val_[0] = (char)(extra+(int)'0');
      
      delete[] temp;
    }
    
  }
  
  else{ //have different sign
 
    if(left.sign_==longInt::pos){
      result = left - (-right);
    }
    else{
      result = right - (-left);
    }
  }
  
  return result;
}

    	
longInt operator- (const longInt& left, const longInt& right)
{
  int i,s;
  int borrow=0;

  longInt result, shortInt;
  int hand_short, hand_result; 


  if(left.sign_==right.sign_){ //both positive or negative
    
    if(left.abs()>=right.abs()){ //left has big absolute value, i.e. longer
      result = left;
      shortInt = right;
    }
    else{
      result = right;
      shortInt = left;
    }
    

    for(i = 0; i<shortInt.size_; i++){
      hand_short = shortInt.size_-1-i;
      hand_result = result.size_-1-i;
      
      s = (int)result.val_[hand_result]-(int)'0'
	-((int)shortInt.val_[hand_short]-(int)'0')-borrow;
      
      if(s<0){
	borrow = 1;
	s += 10;
      }
      else{
	borrow = 0;
      }
     
      result.val_[hand_result] = (char) (s+(int)'0');     
    }
      
    if(borrow>0){

      while(hand_result>0){
	hand_result--;

	s = (int)result.val_[hand_result] -(int)'0'- borrow;

	if(s<0){
	  borrow = 1;
	  s += 10;
	  result.val_[hand_result] = (char) (s+(int)'0');
	}
	else{
	  borrow = 0;
	  result.val_[hand_result] = (char) (s+(int)'0');
	  break;
	}
      }//end while
      
    } //end if (borrow>0){
    
    //decide result's sign
    if(left>=right) result.sign_ = longInt::pos;
    else result.sign_ =longInt::neg;
  }

  else{ //have different sign
    result = left + (-right);
  }
  
  result.reduce();
  return result;
  
  
}


longInt operator* (const longInt& left, const longInt& right)
{

  longInt zero("0");
  if(left==zero || right==zero){
    return zero;
  }

  longInt result;
  int i,j;
  int hand1,hand2 , val1, val2;
  int vl, vr, s, rem, extra;


  result.size_ = left.size_ + right.size_;
  if(left.sign_==right.sign_) result.sign_ = longInt::pos;
  else result.sign_ = longInt::neg;
  
  result.val_ = new char[result.size_+1];
  for(i=0; i<result.size_;i++){
    result.val_[i] = '0';
  }

  
  for(i=0; i<right.size_; i++){
    for(j=0; j<left.size_;j++){

      longInt add(result.size_, result.sign_, '0');

      //hand2 = result.size_ - ((right.size_-i-1)+(left.size_-j-1)+1);
      hand2 = i+j+1;
      hand1 = hand2-1;
      vr = (int)right.val_[i] - (int)'0';
      vl = (int)left.val_[j] - (int)'0';
      s = vr*vl;
     
      extra = s/10;
      rem = s%10;

      val1 = (int)result.val_[hand1] - (int)'0';
      val2 = (int)result.val_[hand2] - (int)'0';
      
      s = extra + val1;
      
      if(s>=10){
	result.val_[hand1] = (char)(s%10+(int)'0');
	add.val_[hand1-1] = (char)(s/10+(int)'0');
      }
      else{
	result.val_[hand1] = (char)(s+(int)'0');
      }
      
      s = rem + val2;
      
      if(s>=10){
	result.val_[hand2] = (char)(s%10+(int)'0');
	add.val_[hand2-1] = (char)(s/10+(int)'0');
      }
      else{
	result.val_[hand2] = (char)(s+(int)'0');
      }
      
      result = result + add;
    }
  }
  
  result.reduce();
  
  return result;
 
}

longInt operator* (const longInt& left, const int& right)
{
}



longInt operator/ (const longInt& left, const longInt& right)
{
  longInt zero("0");
  if(left==zero){
    return zero;
  }
  if(right==zero){
    cerr<<"Error in operator /: Can not divided by 0."<<endl;
    exit(1);
  }
  
  longInt::sign sg;
  if(left.sign_==right.sign_){
    sg = longInt::pos;
  }
  else{
    sg = longInt::neg;
  }
  
  longInt left_abs = left.abs();
  longInt right_abs = right.abs();
  

  if(left_abs<right_abs){
    return zero;
  }
  
  int i;
  int resultSize = left_abs.size_ - right_abs.size_ + 1;
  
  longInt result;
  longInt start(resultSize-1,longInt::pos,'0');
  start.val_[0]='1';
  
  longInt stop(resultSize,longInt::pos,'9');
  longInt mid, diff;
  
  if(resultSize == 1){
    start = "1";
    stop = "9";
  }

 

  //do a binary search for the answer
  while(1){
    if (left_abs.isDivision(right_abs,start)){ //start is the solution
      result = start;
      break;
    }
    else if (left_abs.isDivision(right_abs,stop)){ //stop is the solution
      result = stop;
      break;
    }
    else{ //both are not
      diff = stop - start;
      mid = start + diff.half();
      longInt prod = mid*right_abs;
      
      if (prod>left_abs){
	stop = mid;
      }
      else if(prod<left_abs){
	start = mid;
      }
      else {
	result = mid;
	break;
      }
    }
  }
  
  result.sign_ = sg;
  

  return result;
}



longInt operator% (const longInt& left, const longInt& right)
{
  return left - right*(left/right);
}


	  
  
bool operator< (const longInt& left, const longInt& right)
{
   return !(left>=right);
}

      


bool operator<= (const longInt& left, const longInt& right)
{
  //left.reduce();
  //right.reduce();
  
  
  bool result = true;
  int i;


  if(left.sign_==longInt::pos){
    if(right.sign_==longInt::neg){
      result=false;
    }
    else{
      if(right.size_<left.size_){
	result=false;
      }
      else if(right.size_>left.size_){
	result = true;
      }
      else{//=
	for(i=0; i<right.size_; i++){
	  if(left.val_[i]>right.val_[i]){
	    result = false;
	    break;
	  }
	  else if(left.val_[i]<right.val_[i]){
	    result = true;
	    break;
	  }
	  else{//=, continue
	    ;
	  }  
	}
      }
    }
  }
  
  
  else{ //left.sign_==neg
    if(right.sign_==longInt::neg){
      if(right.size_<left.size_){
	result = true;
      }
      else if(right.size_>left.size_){
	result = false;
      }
      else{//=, compare digit by digit
	for(i=0; i<right.size_; i++){
	  if(left.val_[i]<right.val_[i]){
	    result = false;
	    break;
	  }
	  else if(left.val_[i]>right.val_[i]){
	    result = true;
	    break;
	  }
	  else{//=, continue
	    ;
	  }  
	}
      }
    }
    else{
      result=true;
    }
  }
  
  
  return result;
}


bool operator> (const longInt& left, const longInt& right)
{
  return !(left<=right);
  
}

bool operator>= (const longInt& left, const longInt& right)
{
  
  return ((-left)<=(-right));
  
  
}


bool operator== (const longInt& left, const longInt& right)
{
  if (left.sign_!=right.sign_){
    return false;
  }

  //left.reduce();
  //right.reduce();

  if (left.size_!=right.size_){
    return false;
  }

  bool result = true;

  for(int i=0; i<left.size_; i++){
    if(left.val_[i]!=right.val_[i]){
      result = false;
      break;
    }
  }
  
  return result;
  
}



ostream& operator<< (ostream& out, const longInt& li)
{

  int i;

  if (li.sign_ == longInt::neg) out<<"-";
  
  for(i=0; i<li.size_; i++){
      out<<li.val_[i];
  }
 
  return out;
}

  

istream& operator>> (istream& in, longInt& li)
{
  char* s =new char[1000];
  in>>s;
  
  li = longInt(s);
  return in;
}

    
//private function: remove leading 0s
void longInt::reduce()
{
  
  if(size_<=1) return;
  
  
  longInt result;
  
  int i, cur=0, s = size_;

  while(s>1){
    if(val_[cur]=='0'){
      cur++;
      s--;
    }
    else{
      break;
    }
  }
  
  result.val_ = new char[s+1];
  for(i=0; i< s; i++){
    result.val_[i] = val_[i+cur];
  }
  
  result.sign_ = sign_;
  result.size_ = s;

  (*this) = result;
  
  return;
  

  
}

//for operator /
//a is divisor, hence fixed
//assume all positive
bool longInt::isDivision(const longInt& a, const longInt& b)
{
  bool result = false;
  
  longInt prod = a*b, one("1");
  
  if(prod==(*this)){
    result = true;
  }
  else if(prod>(*this)){
    result = false;
  }
  else{
    if(a*(b+one)>(*this)){
      result = true;
    }
  }
  
  return result;
}

      
  




 



#endif




